#
# coding=utf-8
"""test_week - Test ISO-8601 week handling"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from datetime import date

from expecter import expect
from nose2.tools import params

from rdial.utils import iso_week_to_date


@params(
    (2007, 1, (date(2007, 1, 1), date(2007, 1, 8))),  # Year starts same day
    (2009, 53, (date(2009, 12, 28), date(2010, 1, 4))),  # ISO year spans 2010
    (2013, 52, (date(2013, 12, 23), date(2013, 12, 30))),
)
def test_iso_week_to_date(year, week, expected):
    expect(iso_week_to_date(year, week)) == expected



#
# coding=utf-8
"""rdial - Simple time tracking for simple people"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from . import _version


__version__ = _version.dotted
__date__ = _version.date
__author__ = 'James Rowe <jnrowe@gmail.com>'
__copyright__ = 'Copyright (C) 2011-2014  James Rowe <jnrowe@gmail.com>'
__license__ = 'GNU General Public License Version 3'
__credits__ = ''
__history__ = 'See git repository'

from email.utils import parseaddr

# pylint: disable-msg=W0622
__doc__ += """.

A simple time tracking tool, with no frills and no fizzy coating.

.. moduleauthor:: `%s <mailto:%s>`__
""" % parseaddr(__author__)



#
# coding=utf-8
"""i18n - Internationalisation support for rdial"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# pylint: disable-msg=C0121

import gettext

from os import path

PACKAGE_LOCALE = path.join(path.realpath(path.dirname(__file__)), 'locale')

gettext.install('rdial', PACKAGE_LOCALE)

_, N_ = gettext.gettext, gettext.ngettext



#
# coding=utf-8
"""test_repr - Test __repr__ methods for self-reproducibility"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest

from expecter import expect

from rdial.events import (Event, Events)


def eval_repr(obj):
    """Regenerate an object from its __repr__ output.

    :param object obj: Object to evaluate

    """
    return repr(eval(repr(obj)))


class ReprTest(unittest.TestCase):

    """repr() should return self-documenting string."""

    def test_event(self):
        ev = Event('task', '2011-05-05T11:23:48Z', 'PT01H00M00S')
        expect(repr(ev)) == eval_repr(ev)

    def test_event_no_delta(self):
        ev = Event('task', '2011-05-05T11:23:48Z', '')
        expect(repr(ev)) == eval_repr(ev)

    def test_event_with_message(self):
        ev = Event('task', '2011-05-05T11:23:48Z', 'PT01H00M00S', 'message')
        expect(repr(ev)) == eval_repr(ev)

    def test_events(self):
        ev1 = Event('task', '2011-05-05T11:23:48Z', 'PT01H00M00S')
        ev2 = Event('task', '2011-05-05T12:23:48Z', 'PT00H30M00S')
        events = Events([ev1, ev2])
        expect(repr(events)) == eval_repr(events)



#
# coding=utf-8
"""compat - Python 2/3 compatibility support for rdial"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from sys import version_info

PY2 = version_info[0] == 2

if PY2:
    import codecs

    open = codecs.open
else:
    open = open



#
# coding=utf-8
"""utils - Utility functions for rdial"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# pylint: disable-msg=C0121

import datetime
import functools
import os
import re

try:  # For Python 3
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import SafeConfigParser as ConfigParser  # NOQA

import blessings

from . import compat
from .i18n import _


T = blessings.Terminal()


# Set up informational message functions
def _colourise(text, colour):
    """Colour text, if possible.

    :param str text: Text to colourise
    :param str colour: Colour to display text in
    :rtype: :obj:`str`
    :return str: Colourised text, if possible

    """
    return getattr(T, colour.replace(' ', '_'))(text)


def success(text):
    """Pretty print a success message.

    :param str text: Text to format
    :rtype: :obj:`str`
    :return: Bright green text, if possible

    """
    return _colourise(text, 'bright green')


def fail(text):
    """Pretty print a failure message.

    :param str text: Text to format
    :rtype: :obj:`str`
    :return: Bright red text, if possible

    """
    return _colourise(text, 'bright red')


def warn(text):
    """Pretty print a warning message.

    :param str text: Text to format
    :rtype: ``str``
    :return: Bright yellow text, if possible

    """
    return _colourise(text, 'bright yellow')


class RdialError(ValueError):

    """Generic exception for rdial."""

    if not compat.PY2:
        @property
        def message(self):
            """Compatibility hack for Python 3."""
            return self.args[0]


class UTC(datetime.tzinfo):

    """UTC timezone object."""

    def __repr__(self):
        return 'UTC()'

    # pylint: disable-msg=W0613
    def utcoffset(self, datetime_):
        return datetime.timedelta(0)

    def dst(self, datetime_):
        return datetime.timedelta(0)

    def tzname(self, datetime_):
        return 'UTC'
    # pylint: enable-msg=W0613

utc = UTC()


def parse_delta(string):
    """Parse ISO-8601 duration string.

    :param str string: Duration string to parse
    :rtype: :obj:`datetime.timedelta`
    :return: Parsed delta object

    """
    if not string:
        return datetime.timedelta(0)
    match = re.match("""
        P
        ((?P<days>\d+)D)?
        T?
        ((?P<hours>\d{1,2})H)?
        ((?P<minutes>\d{1,2})M)?
        ((?P<seconds>\d{1,2})?(\.(?P<microseconds>\d+)S)?)
    """, string, re.VERBOSE)
    match_dict = dict((k, int(v) if v else 0)
                      for k, v in match.groupdict().items())
    return datetime.timedelta(**match_dict)  # pylint: disable-msg=W0142


def format_delta(timedelta_):
    """Format ISO-8601 duration string.

    :param datetime.timedelta timedelta_: Duration to process
    :rtype: :obj:`str`
    :return: ISO-8601 representation of duration

    """
    if timedelta_ == datetime.timedelta(0):
        return ''
    days = '%dD' % timedelta_.days if timedelta_.days else ''
    hours, minutes = divmod(timedelta_.seconds, 3600)
    minutes, seconds = divmod(minutes, 60)
    hours = '%02dH' % hours if hours else ''
    minutes = '%02dM' % minutes if minutes else ''
    seconds = '%02dS' % seconds if seconds else ''
    return 'P%s%s%s%s%s' % (days, 'T' if hours or minutes or seconds else '',
                            hours, minutes, seconds)


def parse_datetime(string):
    """Parse ISO-8601 datetime string.

    :param str string: Datetime string to parse
    :rtype: :obj:`datetime.datetime`
    :return: Parsed datetime object

    """
    if not string:
        datetime_ = utcnow()
    else:
        datetime_ = datetime.datetime.strptime(string, '%Y-%m-%dT%H:%M:%SZ')
        datetime_ = datetime_.replace(tzinfo=utc)
    return datetime_


def format_datetime(datetime_):
    """Format ISO-8601 datetime string.

    :param datetime.datetime datetime_: Datetime to process
    :rtype: str
    :return: ISO-8601 compatible string

    """
    # Can't call isoformat method as it uses the +00:00 form
    return datetime_.strftime('%Y-%m-%dT%H:%M:%SZ')


def iso_week_to_date(year, week):
    """Generate date range for a given ISO-8601 week.

    ISO-8601 defines a week as Monday to Sunday, with the first week of a year
    being the first week containing a Thursday.

    :param int year: Year to process
    :param int week: Week number to process
    :rtype: :obj:`tuple` of :obj:`datetime.date`
    :return: Date range objects for given week

    """
    bound = datetime.date(year, 1, 4)
    iso_start = bound - datetime.timedelta(days=bound.isocalendar()[2] - 1)
    start = iso_start + datetime.timedelta(weeks=week - 1)
    end = start + datetime.timedelta(weeks=1)
    return start, end


def utcnow():
    """Wrapper for producing timezone aware current timestamp.

    :rtype: obj:`datetime.datetime`
    :return: Current date and time, in UTC

    """
    return datetime.datetime.utcnow().replace(tzinfo=utc)


def read_config(parser, user_config=None):
    """Read configuration data.

    :type argparse.ArgumentParser parser: Command line parser
    :type str user_config: User defined config file
    :rtype: ConfigParser
    :return: Parsed configuration data

    """
    if user_config and not os.path.exists(user_config):
        raise parser.error(_("Config file %r doesn't exist!" % user_config))

    configs = [os.path.dirname(__file__) + '/config', ]
    for s in os.getenv('XDG_CONFIG_DIRS', '/etc/xdg').split(':'):
        p = s + '/rdial/config'
        if os.path.isfile(p):
            configs.append(p)
    configs.append(xdg_config_location() + '/config')
    configs.append(os.path.abspath('.rdialrc'))
    if user_config:
        configs.append(user_config)
    cfg = ConfigParser()
    for file in configs:
        if os.path.isfile(file):
            cfg.readfp(compat.open(file, encoding='utf-8'))

    return cfg


def write_current(f):
    """Decorator to write ``current`` file on function exit.

    :seealso: :doc:`/taskbars`

    :rtype: obj:`function`

    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        f(*args, **kwargs)
        open('%s/.current' % kwargs['directory'], 'w').write(kwargs['task'])
    return wrapper


def remove_current(f):
    """Decorator to remove ``current`` file on function exit.

    :seealso: :doc:`/taskbars`

    :rtype: obj:`function`

    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        f(*args, **kwargs)
        if os.path.isfile('%s/.current' % kwargs['directory']):
            os.unlink('%s/.current' % kwargs['directory'])
    return wrapper


def xdg_config_location():
    """Return a config location honouring $XDG_CONFIG_HOME.

    :rtype: :obj:`str`

    """
    user_dir = os.getenv('XDG_CONFIG_HOME',
                         os.path.join(os.getenv('HOME', '/'), '.config'))
    return os.path.join(user_dir, 'rdial')


def xdg_data_location():
    """Return a data location honouring $XDG_DATA_HOME.

    :rtype: :obj:`str`

    """
    user_dir = os.getenv('XDG_DATA_HOME', os.path.join(os.getenv('HOME', '/'),
                         '.local/share'))
    return os.path.join(user_dir, 'rdial')



#
# coding=utf-8
"""test_event - Test database handling"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os

from datetime import (datetime, timedelta)
from filecmp import dircmp

from expecter import expect
from nose2.tools import params

from rdial.events import (Event, Events)
from rdial.utils import (parse_datetime, parse_delta, utc)


@params(
    ('test', None, None, None),
    ('test', '2013-02-26T19:45:14Z', None, None),
    ('test', datetime(2013, 2, 26, 19, 45, 14, tzinfo=utc), None,
     None),
    ('test', '2013-02-26T19:45:14Z', 'PT8M19.770869S', None),
    ('test', '2013-02-26T19:45:14Z', timedelta(minutes=8, seconds=19.770869),
     None),
    ('test', '2013-02-26T19:45:14Z', 'PT8M19.770869S', 'stopped'),
)
def test_event_creation(task, start, delta, message):
    e = Event(task, start, delta, message)
    expect(e.task) == task
    if isinstance(start, datetime):
        expect(e.start) == start
    elif start:
        expect(e.start) == parse_datetime(start)
    else:
        # Special case to ignore comparison against utcnow()
        pass
    if isinstance(delta, timedelta):
        expect(e.delta) == delta
    else:
        expect(e.delta) == parse_delta(delta)
    expect(e.message) == message


@params(
    ('test', 3),
    ('date_filtering', 3),
    ('test_not_running', 3),
)
def test_read_datebase(database, events):
    expect(len(Events.read('tests/data/' + database))) == events


@params(
    (0, 'task', datetime(2011, 5, 4, 8, tzinfo=utc),
     timedelta(hours=1)),
    (1, 'task2', datetime(2011, 5, 4, 9, 15, tzinfo=utc),
     timedelta(minutes=15)),
    (2, 'task', datetime(2011, 5, 4, 9, 30, tzinfo=utc), timedelta()),
)
def test_check_events(n, task, start, delta):
    # FIXME: Clean-ish way to perform check, with the caveat that it parses the
    # database on each entry.  Need a better solution.
    events = Events.read('tests/data/test')
    expect(events[n].task) == task
    expect(events[n].start) == start
    expect(events[n].delta) == delta


def test_write_database():
    events = Events.read('tests/data/test')
    events._dirty = events.tasks()
    try:
        events.write('tests/data/test_write')
    except:
        pass
    else:
        comp = dircmp('tests/data/test', 'tests/data/test_write')
        expect(comp.diff_files) == []
        expect(comp.left_only) == []
        expect(comp.right_only) == []
        expect(comp.funny_files) == []
    finally:
        for i in os.listdir('tests/data/test_write'):
            os.unlink('tests/data/test_write/%s' % i)
        os.rmdir('tests/data/test_write')


def test_store_messages_with_events():
    events = Events.read('tests/data/test')
    expect(events.last().message) == 'finished'


def test_non_existing_database():
    expect(Events()) == Events.read('I_NEVER_EXIST')



#
# coding=utf-8
"""test_event_triggers - Test event trigger handling"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


from expecter import expect

from rdial.events import (Events, TaskNotExistError, TaskNotRunningError,
                          TaskRunningError)


def test_start_event():
    events = Events.read('tests/data/test_not_running')
    events.start(task='task2')
    expect(events.running()) == 'task2'


def test_fail_start_when_task_typo():
    events = Events.read('tests/data/test_not_running')
    with expect.raises(TaskNotExistError,
                       "Task non_existant does not exist!  Use `--new' to "
                       'create it'):
        events.start(task='non_existant')


def test_fail_start_when_running():
    events = Events.read('tests/data/test')
    with expect.raises(TaskRunningError, 'Running task task!'):
        events.start(task='task2')


def test_stop_event():
    events = Events.read('tests/data/test')
    events.stop()
    expect(events.running()) is False


def test_stop_event_with_message():
    events = Events.read('tests/data/test')
    events.stop(message='test')
    last = events.last()
    expect(last.message) == 'test'


def test_fail_stop_when_not_running():
    events = Events.read('tests/data/test_not_running')
    with expect.raises(TaskNotRunningError, 'No task running!'):
        events.stop()



#
# coding=utf-8
"""cmdline - Command line functionality for rdial"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# pylint: disable-msg=C0121

import argparse
import datetime
import os
import shlex
import subprocess

try:  # For Python 3
    from configparser import NoOptionError
except ImportError:
    from ConfigParser import NoOptionError  # NOQA

import aaargh
import prettytable

from .events import (Events, TaskRunningError)
from .i18n import (_, N_)
from . import _version
from . import utils

APP = aaargh.App(
    description=_('Simple time tracking for simple people'),
    epilog=_('Please report bugs to https://github.com/JNRowe/issues')
)


# pylint: disable-msg=R0903
class TaskAction(argparse.Action):

    """Define task name, handling --from-dir option."""

    def __call__(self, parser, namespace, values, option_string=None):
        if namespace.task is True:
            namespace.task = os.path.basename(os.path.abspath(os.curdir))
        else:
            namespace.task = values
# pylint: enable-msg=R0903


def task_name_typecheck(string):
    """Check given task name is valid.

    :param str string: Task name to check
    :rtype: :obj:`str`
    :returns: Task name, if valid
    :raises argparse.ArgparseTypeError: If task name is invalid

    """
    if string.startswith('.') or '/' in string or '\000' in string:
        raise argparse.ArgumentTypeError(_('%r is not a valid task name')
                                         % string)
    return string


# pylint: disable-msg=C0103
task_parser = argparse.ArgumentParser(add_help=False)
names_group = task_parser.add_mutually_exclusive_group()
names_group.add_argument('-x', '--from-dir', action='store_true', dest='task',
                         help=_('use directory name as task'))
names_group.add_argument('task', default='default', nargs='?',
                         action=TaskAction, help=_('task name'),
                         type=task_name_typecheck)

duration_parser = argparse.ArgumentParser(add_help=False)
duration_parser.add_argument('-d', '--duration', default='all',
                             choices=['day', 'week', 'month', 'year', 'all'],
                             help=_('filter events for specified time period'))
# pylint: enable-msg=C0103


def filter_events(directory, task=None, duration=None):
    """Filter events for report processing.

    :param str directory: Directory to read events from
    :param str task: Task name to filter on
    :param str duration: Time window to filter on
    :rtype: :obj:`rdial.events.Events`
    :return: Events matching specified criteria

    """
    events = Events.read(directory)
    if task:
        events = events.for_task(task)
    if not duration == 'all':
        if duration == 'week':
            today = datetime.date.today()
            events = events.for_week(*today.isocalendar()[:2])
        else:
            year, month, day = datetime.date.today().timetuple()[:3]
            if duration == 'month':
                day = None
            elif duration == 'year':
                month = None
                day = None
            events = events.for_date(year, month, day)
    return events


def start_time_typecheck(string):
    """Check given start time is valid.

    :param str string: Timestamps to check
    :rtype: :obj:`str`
    :returns: Timestamp, if valid
    :raises argparse.ArgparseTypeError: If timestamp is invalid

    """
    try:
        utils.parse_datetime(string)
    except ValueError:
        raise argparse.ArgumentTypeError(_('%r is not a valid ISO-8601 time '
                                           'string') % string)
    return string


@APP.cmd(help=_('check storage consistency'))
def fsck(directory, backup, config):
    """Check storage consistency.

    :param str directory: Directory to read events from
    :param bool backup: Whether to create backup files
    :param ConfigParser config: Configuration data

    """
    with Events.context(directory, backup) as events:
        last = events[0]
        for event in events[1:]:
            if not last.start + last.delta <= event.start:
                print(utils.fail(_('Overlap:')))
                print(utils.warn('  %r' % last))
                print('  %r' % event)
            last = event


@APP.cmd(help=_('start task'), parents=[task_parser])
@APP.cmd_arg('-n', '--new', action='store_true', help=_('start a new task'))
@APP.cmd_arg('-t', '--time', metavar='time', default='',
             help=_('set start time'), type=start_time_typecheck)
@utils.write_current
def start(directory, backup, config, task, new, time):
    """Start task.

    :param str directory: Directory to read events from
    :param bool backup: Whether to create backup files
    :param ConfigParser config: Configuration data
    :param str task: Task name to operate on
    :param bool new: Create a new task
    :param datetime.datetime time: Task start time

    """
    with Events.context(directory, backup) as events:
        events.start(task, new, time)


@APP.cmd(help=_('stop task'))
@APP.cmd_arg('-m', '--message', metavar='message', help=_('closing message'))
@APP.cmd_arg('-F', '--file', metavar='file', type=argparse.FileType(),
             help=_('read closing message from file'))
@APP.cmd_arg('--amend', action='store_true',
             help=_('amend previous stop entry'))
@utils.remove_current
def stop(directory, backup, config, message, file, amend):
    """Stop task.

    :param str directory: Directory to read events from
    :param bool backup: Whether to create backup files
    :param ConfigParser config: Configuration data
    :param str message: Message to assign to event
    :param str file: Filename to read message from
    :param bool amend: Amend a previously stopped event

    """
    if file:
        message = file.read()
    with Events.context(directory, backup) as events:
        last = events.last()
        if amend and last.running():
            raise TaskRunningError(_("Can't amend running task %s!")
                                   % last.task)
        if amend and not message:
            event = events.last()
            message = event.message
        events.stop(message, force=amend)
    event = events.last()
    print(_('Task %s running for %s') % (event.task,
                                         str(event.delta).split('.')[0]))


@APP.cmd(help=_('switch to another task'), parents=[task_parser])
@APP.cmd_arg('-n', '--new', action='store_true', help=_('start a new task'))
@APP.cmd_arg('-m', '--message', metavar='message',
             help=_('closing message for current task'))
@APP.cmd_arg('-F', '--file', metavar='file', type=argparse.FileType(),
             help=_('read closing message for current task from file'))
@utils.write_current
def switch(directory, backup, config, task, new, message, file):
    """Complete last task and start new one.

    :param str directory: Directory to read events from
    :param bool backup: Whether to create backup files
    :param ConfigParser config: Configuration data
    :param str task: Task name to operate on
    :param bool new: Create a new task
    :param str message: Message to assign to event
    :param str file: Filename to read message from

    """
    if file:
        message = file.read()
    with Events.context(directory, backup) as events:
        if new or task in events.tasks():
            # This is dirty, but we kick on to Events.start() to save
            # duplication of error handling for task names
            events.stop(message)
        event = events.last()
        events.start(task, new)
    print(_('Task %s running for %s') % (event.task,
                                         str(event.delta).split('.')[0]))


@APP.cmd(help=_('run command with timer'), parents=[task_parser])
@APP.cmd_arg('-n', '--new', action='store_true', help=_('start a new task'))
@APP.cmd_arg('-t', '--time', metavar='time', default='',
             help=_('set start time'), type=start_time_typecheck)
@APP.cmd_arg('-m', '--message', metavar='message', help=_('closing message'))
@APP.cmd_arg('-F', '--file', metavar='file', type=argparse.FileType(),
             help=_('read closing message from file'))
@APP.cmd_arg('-c', '--command', metavar='command', help=_('command to run'))
def run(directory, backup, config, task, new, time, message, file, command):
    """Run timed command.

    :param str directory: Directory to read events from
    :param bool backup: Whether to create backup files
    :param ConfigParser config: Configuration data
    :param str task: Task name to operate on
    :param bool new: Create a new task
    :param datetime.datetime time: Task start time
    :param str message: Message to assign to event
    :param str file: Filename to read message from
    :param str command: Command to run

    """
    with Events.context(directory, backup) as events:
        if events.running():
            raise TaskRunningError(_('Task %s is already started!'
                                     % events.last().task))

        try:
            p = subprocess.Popen(command, shell=True)
        except OSError as e:
            raise utils.RdialError(e.strerror)

        events.start(task, new, time)
        open('%s/.current' % directory, 'w').write(task)

        p.wait()

        if file:
            message = file.read()
        events.stop(message)
    event = events.last()
    print(_('Task %s running for %s') % (event.task,
                                         str(event.delta).split('.')[0]))
    os.unlink('%s/.current' % directory)


@APP.cmd(help=_('run predefined command with timer'))
@APP.cmd_arg('-t', '--time', metavar='time', default='',
             help=_('set start time'), type=start_time_typecheck)
@APP.cmd_arg('-m', '--message', metavar='message', help=_('closing message'))
@APP.cmd_arg('-F', '--file', metavar='file', type=argparse.FileType(),
             help=_('read closing message from file'))
@APP.cmd_arg('wrapper', default='default', help=_('wrapper name'))
def wrapper(directory, backup, config, time, message, file, wrapper):
    """Run predefined timed command.

    :param str directory: Directory to read events from
    :param bool backup: Whether to create backup files
    :param ConfigParser config: Configuration data
    :param datetime.datetime time: Task start time
    :param str message: Message to assign to event
    :param str file: Filename to read message from
    :param str wrapper: Run wrapper to execute

    """
    try:
        command = config.get('run wrappers', wrapper)
    except NoOptionError:
        raise ValueError(_('No such wrapper %r') % wrapper)
    parser = argparse.ArgumentParser(parents=[task_parser, ])
    parser.add_argument('-c', '--command')
    args = parser.parse_args(shlex.split(command))
    run(directory, backup, config, args.task, False, time, message, file,
        args.command)


# pylint: disable-msg=C0103
output_parser = argparse.ArgumentParser(add_help=False)
output_group = output_parser.add_mutually_exclusive_group()
output_group.add_argument('--html', action='store_true',
                          help=_('produce HTML output'))
output_group.add_argument('--human', action='store_true',
                          help=_('produce human-readable output'))
# pylint: enable-msg=C0103


@APP.cmd(help=_('report time tracking data'),
         parents=[duration_parser, task_parser, output_parser])
@APP.cmd_arg('-s', '--sort', default='task', choices=['task', 'time'],
             help=_('field to sort by'))
@APP.cmd_arg('-r', '--reverse', action='store_true',
             help=_('reverse sort order'))
def report(directory, backup, config, task, duration, sort, reverse, html,
           human):
    """Report time tracking data.

    :param str directory: Directory to read events from
    :param bool backup: Whether to create backup files
    :param ConfigParser config: Configuration data
    :param str task: Task name to operate on
    :param str duration: Time window to filter on
    :param str sort: Key to sort events on
    :param bool reverse: Reverse sort order
    :param bool html: Produce HTML output
    :param bool human: Produce human-readble output

    """
    if task == 'default':
        # Lazy way to remove duplicate argument definitions
        task = None
    events = filter_events(directory, task, duration)
    if human:
        print(N_('%d event in query', '%d events in query', len(events))
              % len(events))
        print(_('Duration of events %s') % events.sum())
        print(_('First entry started at %s') % events[0].start)
        print(_('Last entry started at %s') % events[-1].start)
        dates = set(e.start.date() for e in events)
        print(_('Events exist on %d dates') % len(dates))
    else:
        table = prettytable.PrettyTable(['task', 'time'])
        formatter = table.get_html_string if html else table.get_string
        try:
            table.align['task'] = 'l'
        except AttributeError:  # prettytable 0.5 compatibility
            table.set_field_align('task', 'l')
        for task in events.tasks():
            table.add_row([task, events.for_task(task).sum()])

        print(formatter(sortby=sort, reversesort=reverse))
    if events.running() and not html:
        current = events.last()
        print(_("Running `%s' since %s")
              % (current.task, utils.format_datetime(current.start)))


@APP.cmd(help=_('display running task, if any'))
def running(directory, backup, config):
    """Display running task, if any.

    :param str directory: Directory to read events from
    :param bool backup: Whether to create backup files
    :param ConfigParser config: Configuration data

    """
    events = Events.read(directory)
    if events.running():
        current = events.last()
        print(_('Task %s has been running for %s')
              % (current.task,
                 str(utils.utcnow() - current.start).split('.')[0]))
    else:
        print(utils.warn(_('No task is running!')))


@APP.cmd(help=_('display last event, if any'))
def last(directory, backup, config):
    """Display last event, if any.

    :param str directory: Directory to read events from
    :param bool backup: Whether to create backup files
    :param ConfigParser config: Configuration data

    """
    events = Events.read(directory)
    event = events.last()
    if not events.running():
        print(_('Last task %s, ran for %s') % (event.task, event.delta))
    else:
        print(utils.warn(_('Task %s is still running') % event.task))


@APP.cmd(help=_('generate ledger compatible data file'),
         parents=[duration_parser, task_parser])
@APP.cmd_arg('-r', '--rate', metavar='rate',
             help=_('hourly rate for task output'))
def ledger(directory, backup, config, task, duration, rate):
    """Generate ledger compatible data file.

    :param str directory: Directory to read events from
    :param bool backup: Whether to create backup files
    :param ConfigParser config: Configuration data
    :param str task: Task name to operate on
    :param str duration: Time window to filter on
    :param str rate: Rate to assign hours in report

    """
    if task == 'default':
        # Lazy way to remove duplicate argument definitions
        task = None
    events = filter_events(directory, task, duration)
    if events.running():
        print(_(';; Running event not included in output!'))
    for event in events:
        if not event.delta:
            continue
        end = event.start + event.delta
        # Can't use timedelta.total_seconds() as it was only added in 2.7
        seconds = event.delta.days * 86400 + event.delta.seconds
        hours = seconds / 3600.0
        print('%s-%s' % (event.start.strftime('%Y-%m-%d * %H:%M'),
                         end.strftime('%H:%M')))
        print('    (task:%s)  %.2fh%s'
              % (event.task, hours, ' @ %s' % rate if rate else ''))
    if events.running():
        print(_(';; Running event not included in output!'))


def main():
    """Main script."""
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('--config')
    args, remaining = parser.parse_known_args()
    cfg = utils.read_config(parser, args.config)

    if not cfg.getboolean('rdial', 'colour') or os.getenv('NO_COLOUR'):
        utils._colourise = lambda s, colour: s

    for name, parser in APP._subparsers.choices.items():
        if cfg.has_section(name):
            d = {}
            for k in cfg.options(name):
                try:
                    d[k] = cfg.getboolean(name, k)
                except ValueError:
                    d[k] = cfg.get(name, k)
            parser.set_defaults(**d)

    APP.arg('--version', action='version',
            version='%%(prog)s %s' % _version.dotted)
    APP.arg('-d', '--directory', metavar='dir',
            help=_('directory to read/write to'))
    APP.arg('--no-backup', dest='backup', action='store_true',
            help=_('do not write data file backups'))
    APP.arg('--config', metavar='file',
            help=_('file to read configuration data from'))
    APP.defaults(backup=not cfg.getboolean('rdial', 'backup'),
                 directory=cfg.get('rdial', 'directory',
                                   vars={'xdg_data_location':
                                         utils.xdg_data_location()}),
                 config=cfg)
    try:
        APP.run(remaining)
    except utils.RdialError as error:
        print(utils.fail(error.message))
        return 2



#! /usr/bin/env python
# coding=utf-8
"""setup.py - Setuptools tasks and config for rdial"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from sys import version_info

from setuptools import setup

from rdial import _version


def parse_requires(file):
    deps = []
    req_file = open('extra/%s' % file)
    entries = map(str.strip, req_file.readlines())
    for dep in entries:
        if dep.startswith('-r '):
            deps.extend(parse_requires(dep.split()[1]))
        else:
            deps.append(dep)
    return deps

install_requires = parse_requires('requirements-py%s%s.txt' % version_info[:2])

setup(
    name='rdial',
    version=_version.dotted,
    description='Simple time tracking for simple people',
    long_description=open('README.rst').read(),
    author='James Rowe',
    author_email='jnrowe@gmail.com',
    url='https://github.com/JNRowe/rdial',
    license='GPL-3',
    keywords='timetracking task report',
    packages=['rdial', ],
    include_package_data=True,
    package_data={'': ['config', 'rdial/locale/*/LC_MESSAGES/*.mo']},
    entry_points={'console_scripts': ['rdial = rdial.cmdline:main', ]},
    install_requires=install_requires,
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Other Audience',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Office/Business',
        'Topic :: Office/Business :: News/Diary',
        'Topic :: Office/Business :: Scheduling',
        'Topic :: Other/Nonlisted Topic',
        'Topic :: Utilities',
    ],
)



#
# coding=utf-8
"""test_event_query - Test event query handling"""
# Copyright © 2011-2014  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from datetime import timedelta

from expecter import expect

from rdial.events import Events


def test_list_tasks():
    events = Events.read('tests/data/test')
    expect(events.tasks()) == ['task', 'task2']


def test_current_running_event():
    events = Events.read('tests/data/test')
    expect(events.running()) == 'task'


def test_no_currently_running_event():
    events = Events.read('tests/data/test_not_running')
    expect(events.running()) is False


def test_sum_durations_in_database():
    events = Events.read('tests/data/test_not_running')
    expect(events.sum()) == timedelta(hours=2, minutes=15)
